"""Curses window object."""

import curses


def create_window(
    x: int, y: int, begin_x: int = 0, begin_y: int = 0, init: bool = False
):
    """Initialize curses, make and return window."""

    if init:
        curses.initscr()
        curses.start_color()

    screen = curses.newwin(y, x, begin_x, begin_y)
    screen.keypad(0)
    screen.nodelay(1)

    if init:
        curses.noecho()
        curses.cbreak()
        try:
            curses.curs_set(0)
        except curses.error:
            # old terminals may have no cursor modes support
            pass

    return screen


def deinit_window(screen, deinit: bool = True):
    """Destroy window, deinit curses, make console changes back."""

    screen.nodelay(0)
    screen.keypad(0)

    if deinit:
        curses.nocbreak()
        curses.echo()
        try:
            curses.curs_set(1)
        except curses.error:
            # old terminals may have no cursor modes support
            pass

        curses.endwin()



"""Curses color manipulation objects."""

import curses

from typing import List

from eaf.errors import Error


class PairAlreadyRegistered(Error):
    def __init__(self, attr, value):
        super().__init__(
            f"Pair with {attr} = {value} already registered."
            " You can use force=True to redefine it."
        )


class Pair:
    """Color pair representation."""

    def __init__(self, idx, fg, bg, attrs=None):
        self.idx = idx
        self.bg = bg
        self.fg = fg
        self.attrs = attrs

    def init(self):
        """Init curses color pair.

        Does nothing if terminal doesn't support colors.
        """

        if not curses.has_colors():
            return

        curses.init_pair(self.idx, self.fg, self.bg)


class Palette:
    """Palette abstraction for curses color pairs.

    Has some predefined constants (taken directly from curses),

    Pass to init initial palette as list of tuples of 3 elements: name,
    foreground color, background color:

    palette = Palette([
        ("normal", Palette.COLOR_WHITE, Palette.COLOR_BLACK),
        ("error", Palette.COLOR_RED, Palette.COLOR_BLACK),
    ])
    """

    COLOR_BLACK = curses.COLOR_BLACK
    COLOR_BLUE = curses.COLOR_BLUE
    COLOR_CYAN = curses.COLOR_CYAN
    COLOR_GREEN = curses.COLOR_GREEN
    COLOR_MAGENTA = curses.COLOR_MAGENTA
    COLOR_RED = curses.COLOR_RED
    COLOR_WHITE = curses.COLOR_WHITE
    COLOR_YELLOW = curses.COLOR_YELLOW

    A_BOLD = curses.A_BOLD

    def __init__(self, palette=None):
        self.palette = {}

        idx = 1
        for name, fg, bg in palette or []:
            self.add_pair(name, Pair(idx, fg, bg), init=False)
            idx += 1

    def add_pair(self, name: str, pair: Pair, force: bool = False, init: bool = True):
        """Add new color pair. Use force to override name and idx."""

        if name in self.palette and not force:
            raise PairAlreadyRegistered("name", name)

        if pair.idx in [pair.idx for pair in self.palette.values()] and not force:
            raise PairAlreadyRegistered("idx", pair.idx)

        self.palette[name] = pair

        if init:
            pair.init()

    def init_colors(self):
        """Init all color pairs."""

        for pair in self.palette.values():
            pair.init()

    @property
    def pair_names(self) -> List[str]:
        return sorted(self.palette, key=lambda it: self[it].idx)

    def __getattr__(self, name: str) -> int:
        return self.palette[name].idx



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xo1']

package_data = \
{'': ['*']}

install_requires = \
['eaf>=0.2,<0.3']

setup_kwargs = {
    'name': 'xo1',
    'version': '0.1.2',
    'description': 'EAF-based framework for creating terminal applications',
    'long_description': '|PyPI| |Build Status| |codecov.io|\n\n===\nxo1\n===\n\nPython framework for creating terminal applications.\n\nRequirements\n============\n\n* >=python-3.7\n* >=eaf-0.2\n\nInstallation\n============\n\n.. code-block:: console\n\n\t$ pip install xo1\n\n\nDevelopment\n===========\n\nInstallation\n------------\n\n.. code-block:: console\n\n   $ poetry install\n\nTesting\n-------\n\n.. code-block:: console\n\n   $ poetry run pytest -s -v tests/  # run all tests\n   $ poetry run pytest --cov=xo1 -s -v tests/  # run all tests with coverage\n   $ poetry run black xo1/ tests/  # autoformat code\n   $ # run type checking\n   $ poetry run pytest --mypy --mypy-ignore-missing-imports -s -v xo1/ tests/\n   $ # run code linting\n   $ poetry run pytest --pylint -s -v xo1/ tests/\n\nDocumentation\n-------------\n\n* **To be added**\n\n.. |PyPI| image:: https://badge.fury.io/py/xo1.svg\n   :target: https://badge.fury.io/py/xo1\n.. |Build Status| image:: https://github.com/pkulev/xo1/workflows/CI/badge.svg\n.. |codecov.io| image:: http://codecov.io/github/pkulev/xo1/coverage.svg?branch=master\n   :target: http://codecov.io/github/pkulev/xo1?branch=master\n',
    'author': 'Pavel Kulyov',
    'author_email': 'kulyov.pavel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pkulev/xo1',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""EAF application implementation with curses support."""

from typing import Optional

import eaf.app

from xo1.window import create_window, deinit_window
from xo1.color import Palette
from xo1.render import Renderer


class Application(eaf.app.Application):
    """Curses-powered application class."""

    def __init__(
        self, x, y, palette: Optional[Palette] = None, title: str = "xo1 application"
    ):
        window = create_window(x, y, init=True)
        renderer = Renderer(window)

        super().__init__(renderer, window)

        self._palette = palette or Palette()
        self._palette.init_colors()

        self.set_caption(title)

    def set_caption(self, caption):
        """Set window caption.

        This is very hacky workaround, curses doesn't know that terminal state
        changed. Not all terminals support this.

        .. NOTE:: application doesn't revert title back, maybe later will be
                  implemented.
        """

        print(f"\x1b]0;{caption}\x07")

    def stop(self):
        self._ioloop.add_callback(lambda: deinit_window(self.renderer.screen))
        super().stop()

    @property
    def palette(self) -> Palette:
        """Palette getter."""

        return self._palette

    @palette.setter
    def palette(self, palette: Palette):
        """Palette setter."""

        if not isinstance(palette, Palette):
            raise TypeError(f"Expected Palette class, got {type(palette)}")

        self.palette = palette



"""EAF Renderer implementation for curses graphics backend."""

from operator import attrgetter

import eaf

from eaf.core import Vec3


class Renderer(eaf.Renderer):
    """Curses renderer implementation."""

    INVISIBLE_SYMBOLS = " "
    """Symbols that renderer must not render on the screen."""

    def __init__(self, screen):
        super().__init__(screen)

    def clear(self):
        self.screen.erase()
        self.screen.border(0)

    def render_objects(self, objects):
        """Render all renderable objects."""

        # Window border, rendering behind it will cause curses.error
        border = Vec3(*self.screen.getmaxyx()[::-1])

        # TODO: Move sorting to some kind of object manager
        # Must be sorted on adding objects
        for obj in sorted(objects, key=attrgetter("render_priority")):
            gpos_list, data_gen = obj.get_render_data()

            for data in data_gen:
                for gpos in gpos_list:
                    lpos, image, style = data
                    cpos = (gpos + lpos)[int]

                    if (
                        (cpos.x >= border.x - 1 or cpos.y >= border.y - 1)
                        or (cpos.x <= 0 or cpos.y <= 0)
                    ) and not obj.draw_on_border:
                        # TODO: remove
                        obj.remove_obsolete(gpos)
                        continue

                    if image in self.INVISIBLE_SYMBOLS:
                        continue

                    if style:
                        self.screen.addstr(cpos.y, cpos.x, image, style)
                    else:
                        self.screen.addstr(cpos.y, cpos.x, image)

    def present(self):
        self.screen.refresh()

    def get_width(self):
        return self.screen.getmaxyx()[0]

    def get_height(self):
        return self.screen.getmaxyx()[1]
